/*
 * Copyright (c) 2000, 2001
 *	Gesellschaft fuer wissenschaftliche
 *      Datenverarbeitung mbH Goettingen.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Gesellschaft
 *	fuer wissenschaftliche Datenverarbeitung mbH Goettingen.
 * 4. The name of the Gesellschaft fuer wissenschaftliche Datenverarbeitung
 *    mbH Goettingen may not be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE GESELLSCHAFT FUER WISSENSCHAFTLICHE
 * DATENVERARBEITUNG MBH GOETTINGEN ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE GESELLSCHAFT FUER WISSENSCHAFTLICHE DATENVERARBEITUNG
 * MBH GOETTINGEN BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] =
"@(#) Copyright (c) 2000, 2001\n\
	Gesellschaft fuer wissenschaftliche\n\
        Datenverarbeitung mbH Goettingen.  All rights reserved.\n";
static const char rcsid[] = "$Id: rmonitor.c,v 1.2 2001/07/20 15:30:01 kheuer Exp $";
#endif /* not lint */

/*
 * Remote host monitoring utility.
 *
 * The code has been developed for use on hardware driven by the FreeBSD
 * operating system. Portability to other operating systems has not been
 * an important design goal. Nevertheless, ports to operating systems
 * providing the typical BSD API may be possible without too much effort.
 *
 * Compile, link and install commands:
 *
 *   cc -O rmonitor.c -lncurses -o rmonitor -s
 *
 * 2000/12/14 coding begins
 * 2000/12/22 first usable version
 * 2000/12/29 some improvements
 * 2001/01/02 minor bug fix
 * 2001/03/01 modified output for pre FreeBSD 4.2-RELEASE
 * 2001/06/22 RCS version 1.0 check in
 * 2001/06/22 bug fix (short hostname may be longer than 8 characters now)
 * 2001/06/22 RCS version 1.1 check in
 * 2001/07/20 multicast extension based on patch contributed by
 *            Alain Thivillon <at@teaser.net> and
 *            further improvements, among other things logging
 */


#include <ncurses.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/sysctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/user.h>
#include <sys/vmmeter.h>
#include <time.h>
#include <unistd.h>
#include <utmp.h>


#define DEFTOL      2.5				  /* default tolerance */
#define DEFPORT    3661				  /* default port # */
#define DEFRTIME     30				  /* default refresh time */
#define NOBODY "nobody"				  /* unprivileged user */
#define STRLEN      256				  /* max string length */

						  /* macro function */
#define errmsg(s) \
( \
fprintf(stderr, "%s: %s: %s\n", pgm, s, strerror(errno)) \
)


typedef struct list {
  char        host[STRLEN];
  time_t      lrefr;
  struct list *next;
} list_t;  


void dspstat  ( char * );			  /* display statistics */
void fmtdelt  ( long, char * );			  /* format time difference */
void getstat  ( char * );			  /* get host statistics */
void logstat  ( char * );			  /* log statistics to file */
void onexit   ( void );				  /* exit function */
int  udpclt   ( void );				  /* udp client code */
int  udpsrv   ( void );				  /* udp server code */
void usage    ( int );				  /* print usage note */


char   lhost[STRLEN];				  /* local host name */
char   lfname[MAXPATHLEN + 1];			  /* log file name */
char   netadd[STRLEN];				  /* network address */
char   *pgm;					  /* program name */
FILE   *lf     = NULL;				  /* log file */
float  tol     = DEFTOL;			  /* tolerance */
int    cf      = 0;				  /* client flag */
int    clt     = -1;				  /* client socket */
int    port    = DEFPORT;			  /* port number */
int    quiet   = 0;				  /* quiet mode flag */
int    rtime   = DEFRTIME;			  /* refresh time */
int    srv     = -1;				  /* server socket */
int    ttl     =  4;				  /* default multicast ttl */
list_t *hlist  = NULL;				  /* host list */
WINDOW *win    = NULL;				  /* curses window */


void dspstat ( char *stat )			  /* ----- dspstat ----- */
{
  char   buf[STRLEN];
  char   *pos;
  char   tmp[STRLEN];
  int    y;
  long   delta;
  list_t *curr;
  list_t *found = NULL;
  list_t *hlptr;
  list_t *next;
  list_t *save;
  time_t ct;
  static char *head = "HOST      REFRESH   UPTIME  #U SL  DB   LOADAVG   "
    "PROCESSES   FILES    MEM(ABW)";

#define begx win->_begx
#define begy win->_begy
#define maxx win->_maxx
#define maxy win->_maxy

  if (!win) {
    if (!(win = initscr())) {
      errmsg("initscr");
      exit(EX_UNAVAILABLE);
    }
    clear();
  }

  if ((ct = time(NULL)) == (time_t) -1) errmsg("time");
  strncpy(tmp, ctime(&ct), STRLEN - 1);
  if (pos = strchr(tmp, '\n')) *pos = '\0';
  snprintf(buf, STRLEN, "%s: %s on %s: listening on port %d",
	   tmp, pgm, lhost, port);
  buf[maxx] = '\0';

  move(begy, begx);
  clrtoeol();
  attron(A_STANDOUT);
  mvaddstr(begy, begx + (maxx - strlen(buf)) / 2, buf);
  attroff(A_STANDOUT);

  move(begy + 1, begx);
  clrtoeol();
  attron(A_UNDERLINE);
  mvaddstr(begy + 1, begx + (maxx - strlen(head)) / 2, head);
  attroff(A_UNDERLINE);

  memset(buf, 0, STRLEN);
  strncpy(buf, stat, STRLEN - 1);
  if (pos = strchr(buf, ' ')) *pos = '\0';

  for (hlptr = hlist; hlptr; hlptr = hlptr->next)
    if (!strncmp(hlptr->host, buf, strlen(buf))) {
      strncpy(hlptr->host, stat, STRLEN - 1);
      hlptr->lrefr = ct;
      found = hlptr;
      break;
    }

  if (!found) {

    if (!(save = malloc(sizeof(list_t)))) {
      errmsg("malloc");
      exit(EX_UNAVAILABLE);
    }

    strncpy(save->host, stat, STRLEN - 1);
    save->lrefr = ct;
    save->next  = NULL;

    if (hlist) {
      
      if (strncmp(hlist->host, buf, strlen(buf)) > 0) {
	curr  = hlist;
	hlist = save;
	hlist->next = curr;
	goto display;
      }
      
      for (hlptr = hlist; hlptr; hlptr = hlptr->next) {
	curr = hlptr;
	next = hlptr->next;
	if (next) {
	  if (strncmp(next->host, buf, strlen(buf)) > 0) {
	    save->next = next;
	    curr->next = save;
	    goto display;
	  }
	}
	else {
	  curr->next = save;
	  goto display;
	}
      }
    }
    else
      hlist = save;
  }

 display:

  y = begy + 1;
  for (hlptr = hlist; hlptr && ++y <= maxy; hlptr = hlptr->next) {
    move(y, begx);
    clrtoeol();
    fmtdelt(delta = - (long) (ct - hlptr->lrefr), buf);
    memcpy(hlptr->host + 8, buf, strlen(buf));
    if ((float) abs(delta) > tol * rtime) attron(A_STANDOUT);
    mvaddstr(y, begx + (maxx - strlen(stat)) / 2, hlptr->host);
    if ((float) abs(delta) > tol * rtime) attroff(A_STANDOUT);
  }

  clrtobot();
  move(maxy, begx);
  refresh();

#undef begx
#undef begy
#undef maxx
#undef maxy
}


void fmtdelt ( long delta, char *print )	  /* ----- fmtdelt ----- */
{
  int days;
  int hrs;
  int mins;
  int secs;
  int sign;

  sign  = delta < 0 ? -1 : +1;
  delta = abs(delta);
  days  = delta / 86400;
  hrs   = (delta - 86400 * days) / 3600;
  mins  = (delta - 86400 * days - 3600 * hrs) / 60;
  secs  = delta - 86400 * days - 3600 * hrs - 60 * mins;

  if (cf)
    snprintf(print, STRLEN, "%+4d%c%02d:%02d",
	     sign * days, sign < 0 ? '-' : '+', hrs, mins);
  else
    snprintf(print, STRLEN, "%c%d:%02d:%02d",
	     sign < 0 ? '-' : '+', hrs + 24 * days, mins, secs);
}


void getstat ( char *stat )			  /* ----- getstat ----- */
{
  static char   lhost8[8];
  static char   uptime[STRLEN];
  static int    fstcall = 1;
#define samples 2
  double lavg[samples];
  int    dbufs  = 0;
  int    len;
  int    maxf   = -1;
  int    maxp   = -1;
  int    memfre = 0;
  int    memtot = 0;
  int    memuse = 0;
  int    nu     = -1;
  int    np     = -1;
  int    openf  = -1;
  int    pgcnt  = 0;
  int    pgfre  = 0;
  int    pgiact = 0;
  int    pgsize = 0;
  int    slvl   = 0;
  int    vn[3];
  int    utfd;
  time_t ct;
  struct utmp utmprec;

  union {
    char   buf[STRLEN];
    int    dbufs;
    int    maxf;
    int    maxp;
    int    openf;
    int    pgcnt;
    int    pgfre;
    int    pgiact;
    int    pgsize;
    int    slvl;
    struct timeval bt;
  } loc;

  if (fstcall) {
    fstcall = 0;
    memset(lhost8, 0, 8);
    strncpy(lhost8, lhost, 7);
  }

  memset(lavg, 0, sizeof lavg);
  memset(stat, 0, STRLEN);
  if ((ct = time(NULL)) == (time_t) -1) errmsg("time");

  len   = STRLEN;
  vn[0] = CTL_KERN;
  vn[1] = KERN_BOOTTIME;
  if (sysctl(vn, 2, (void *) &loc, &len, NULL, 0) < 0) errmsg("sysctl");
  fmtdelt((long) ct - (long) loc.bt.tv_sec, uptime);
  
  len   = STRLEN;
  vn[0] = CTL_KERN;
  vn[1] = KERN_SECURELVL;
  if (sysctl(vn, 2, (void *) &loc, &len, NULL, 0) < 0) errmsg("sysctl");
  slvl  = loc.slvl;

  len   = STRLEN;
  vn[0] = CTL_KERN;
  vn[1] = KERN_MAXFILES;
  if (sysctl(vn, 2, (void *) &loc, &len, NULL, 0) < 0) errmsg("sysctl");
  maxf  = loc.maxf;

#if __FreeBSD_version >= 420000
  len   = STRLEN;				  /* cannot use sysctl here */
  if (sysctlbyname("kern.openfiles", (void *) &loc, &len, NULL, 0) < 0)
    errmsg("sysctl");
  openf = loc.openf;
#endif /* __FreeBSD_version >= 420000 */

  len   = STRLEN;
  vn[0] = CTL_KERN;
  vn[1] = KERN_MAXPROC;
  if (sysctl(vn, 2, (void *) &loc, &len, NULL, 0) < 0) errmsg("sysctl");
  maxp  = loc.maxp;

  len   = STRLEN;
  vn[0] = CTL_KERN;
  vn[1] = KERN_PROC;
  vn[2] = KERN_PROC_ALL;
  if (sysctl(vn, 3, NULL, &len, NULL, 0) < 0) errmsg("sysctl");
  np    = len / sizeof(struct kinfo_proc);

  len   = STRLEN;				  /* cannot use sysctl here */
  if (sysctlbyname("vfs.numdirtybuffers", (void *) &loc, &len, NULL, 0) < 0)
    errmsg("sysctl");
  dbufs = loc.dbufs;

  len    = STRLEN;				  /* cannot use sysctl here */
  if (sysctlbyname("vm.stats.vm.v_page_count",
		   (void *) &loc, &len, NULL, 0) < 0) errmsg("sysctl");
  pgcnt  = loc.pgcnt;
  len    = STRLEN;				  /* cannot use sysctl here */
  if (sysctlbyname("vm.stats.vm.v_free_count",
		   (void *) &loc, &len, NULL, 0) < 0) errmsg("sysctl");
  pgfre  = loc.pgfre;
  len    = STRLEN;				  /* cannot use sysctl here */
  if (sysctlbyname("vm.stats.vm.v_inactive_count",
		   (void *) &loc, &len, NULL, 0) < 0) errmsg("sysctl");
  pgiact = loc.pgiact;
  len    = STRLEN;				  /* cannot use sysctl here */
  if (sysctlbyname("vm.stats.vm.v_page_size",
		   (void *) &loc, &len, NULL, 0) < 0) errmsg("sysctl");
  pgsize = loc.pgsize;
  memfre = ((pgsize / 1024) * (pgfre + pgiact)) / 1024;
  memtot = ((pgsize / 1024) * pgcnt) / 1024;
  memuse = memtot - memfre;

  if (getloadavg(lavg, samples) != samples) errmsg("getloadavg");

  if ((utfd = open(_PATH_UTMP, O_RDONLY)) >= 0) {
    nu = 0;
    while (read(utfd, &utmprec, sizeof utmprec) > 0)
      if (*(utmprec.ut_name)) nu++;
    if (close(utfd) < 0) errmsg("close");
  }
  else
    errmsg("open");

#if __FreeBSD_version >= 420000
  snprintf(stat, STRLEN,
	   "%-8s         %s %2d %2d %3d %5.2f %5.2f "
	   "%5d%3.0f%% %5d%3.0f%% %4dM%3.0f%%",
	   lhost8, uptime, nu, slvl, dbufs, lavg[0], lavg[1],
	   np, (100.0 * np) / (float) maxp,
	   openf, (100.0 * openf) / (float) maxf,
	   memuse, (100.0 * memuse) / (float) memtot);
#else
  snprintf(stat, STRLEN,
	   "%-8s         %s %2d %2d %3d %5.2f %5.2f "
	   "%5d%3.0f%%     %c  %c%% %4dM%3.0f%%",
	   lhost8, uptime, nu, slvl, dbufs, lavg[0], lavg[1],
	   np, (100.0 * np) / (float) maxp, '?', '?',
	   memuse, (100.0 * memuse) / (float) memtot);
#endif /* __FreeBSD_version >= 420000 */
#undef samples
}

						  /* ----- logstat ----- */
void logstat ( char *stat )
{
  char   buf[STRLEN];
  time_t ct;
  struct tm *lt;

  if (!lf)
    if (!(lf = fopen(lfname, "a"))) {
      errmsg("fopen");
      exit(EX_UNAVAILABLE);
    }

  memset(buf, 0, STRLEN);
  strncpy(buf, stat, STRLEN - 1);

  if ((ct = time(NULL)) == (time_t) -1) errmsg("time");
  lt = localtime(&ct);
  *(buf + 8) = '\0';
  if (!fprintf(lf, "%04d%02d%02d%02d%02d%02d %s%s\n",
	       lt->tm_year + 1900, lt->tm_mon + 1, lt->tm_mday,
	       lt->tm_hour, lt->tm_min, lt->tm_sec,
	       buf, buf + 18)) errmsg("fprintf");
  if (fflush(lf)) errmsg("fflush");
}


void onexit ( void )				  /* ----- onexit ----- */
{
  if (win)      endwin();
  if (clt >= 0) close(clt);
  if (srv >= 0) close(srv);
  if (lf)       fclose(lf);
}


int udpclt ( void )				  /* ----- udpclt ----- */
{
  char   buf[STRLEN];
  int    bc = 1;
  struct sockaddr_in ca;
  struct sockaddr_in sa;
  unsigned long      na;

  memset(&ca, 0, sizeof ca);
  memset(&sa, 0, sizeof sa);

  if (*netadd) {
    if ((na = inet_addr(netadd)) == INADDR_NONE) {
      errmsg("inet_addr");
      exit(EX_USAGE);
    }
  } else
    na = htonl(INADDR_BROADCAST);

  ca.sin_addr.s_addr = htonl(INADDR_ANY);
  ca.sin_family      = PF_INET;
  ca.sin_port        = htons(0);

  sa.sin_addr.s_addr = na;
  sa.sin_family      = PF_INET;
  sa.sin_port        = htons(port);
  
  if ((clt = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
    errmsg("socket");
    return EX_UNAVAILABLE;
  }

  if (IN_MULTICAST(ntohl(na))) {
    if (setsockopt(clt, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof ttl) < 0) {
      errmsg("setsockopt IP_MULTICAST_TTL");
      return EX_UNAVAILABLE;
    }
  } else {
    if (setsockopt(clt, SOL_SOCKET, SO_BROADCAST, &bc, sizeof bc) < 0) {
      errmsg("setsockopt SO_BROADCAST");
      return EX_UNAVAILABLE;
    }
  }

  if (bind(clt, (struct sockaddr *) &ca, sizeof ca) < 0) {
    errmsg("bind");
    return EX_UNAVAILABLE;
  }

  while (1) {
    getstat(buf);
    sendto(clt, buf, strlen(buf) + 1, 0, (struct sockaddr *) &sa, sizeof sa);
    sleep(rtime);
  }

  return EX_OK;					  /* never reached */
}


int udpsrv ( void )				  /* ----- udpsrv ----- */
{
  char   buf[STRLEN];
  int    bc = 1;
  struct ip_mreq     mreq;
  struct sockaddr_in sa;
  struct sockaddr_in samcast;
  unsigned long      na;

  memset(&sa, 0, sizeof sa);

  sa.sin_addr.s_addr = htonl(INADDR_ANY);
  sa.sin_family      = PF_INET;
  sa.sin_port        = htons(port);
  
  if ((srv = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
    errmsg("socket");
    return EX_UNAVAILABLE;
  }

  if (setsockopt(srv, SOL_SOCKET, SO_REUSEPORT, &bc, sizeof bc) < 0) {
    errmsg("setsockopt SO_REUSEPORT");
    return EX_UNAVAILABLE;
  }

  if (*netadd) {
    if ((na = inet_addr(netadd)) == INADDR_NONE) {
      errmsg("inet_addr");
      exit(EX_USAGE);
    }
    if (IN_MULTICAST(ntohl(na))) {
      memset(&mreq, 0, sizeof mreq);
      mreq.imr_multiaddr.s_addr = na;
      mreq.imr_interface.s_addr = htonl(INADDR_ANY);
      if (setsockopt(srv, IPPROTO_IP, IP_ADD_MEMBERSHIP,
		     &mreq, sizeof mreq) < 0) {
	errmsg("setsockopt IP_ADD_MEMBERSHIP");
	return EX_UNAVAILABLE;
      }
    }
  }

  if (bind(srv, (struct sockaddr *) &sa, sizeof sa) < 0) {
    errmsg("bind");
    return EX_UNAVAILABLE;
  }

  while (1) {

    memset(buf, 0, STRLEN);
    if (recvfrom(srv, buf, sizeof buf, 0, NULL, 0) < 0) {
      if (errno != EINTR) errmsg("recvfrom");
      continue;
    }
    if (!quiet)  dspstat(buf);
    if (*lfname) logstat(buf);
  }

  return EX_OK;					  /* never reached */
}


void usage ( int help )				  /* ----- usage ----- */
{
  char buf[STRLEN];
  int  rc;

  if (!help)
    fprintf(stderr,
	    "Usage: %s [-c] [-h] [-l logfile] [-n address] [-p port] "
	    "[-q] [-r time] [-t ttl] [-w tolerance]\n", pgm);
  else {
    snprintf(buf, STRLEN, "man %s", pgm);
    if (rc = system(buf)) {
      fprintf(stderr, "%s: cannot execute: %s\n");
      exit(rc);
    }
  }

  exit(EX_USAGE);
}


int main ( int argc, char *argv[] )		  /* ----- main ----- */
{
  char   *pos;
  int    ch;
  pid_t  pid;
  struct passwd *pw;

  if (pgm = strrchr(argv[0], '/'))
    pgm++;
  else
    pgm = argv[0];

  if (getuid() == 0) {
    if (!(pw = getpwnam(NOBODY))) {
      errmsg("getpwnam");
      return EX_UNAVAILABLE;
    }
    if (setgid(pw->pw_gid) < 0) {
      errmsg("setgid");
      return EX_UNAVAILABLE;
    }
    if (setuid(pw->pw_uid) < 0) {
      errmsg("setuid");
      return EX_UNAVAILABLE;
    }
  }

  memset(lfname, 0, MAXPATHLEN + 1);
  memset(netadd, 0, STRLEN);

  opterr = 0;
  while ((ch = getopt(argc, argv, ":chl:n:p:qr:t:w:")) != -1) {

    switch (ch) {

    case 'c':
      cf = 1;
      break;

    case 'h':
      usage(1);
      break;

    case 'l':
      strncpy(lfname, optarg, MAXPATHLEN);
      break;

    case 'n':
      strncpy(netadd, optarg, STRLEN - 1);
      break;

    case 'p':
      if (!(port = abs(atoi(optarg)))) usage(0);
      break;

    case 'q':
      quiet = 1;
      break;

    case 'r':
      if (!(rtime = abs(atoi(optarg)))) usage(0);
      break;

    case 't':
      if (!(ttl = abs(atoi(optarg)))) usage(0);
      break;

    case 'w':
      if (!(tol = abs(atof(optarg)))) usage(0);
      if (tol <= 1.0) usage(0);
      break;

    case ':':
    case '?':
    default:
      usage(0);
      break;
    }
  }

  memset(lhost, 0, STRLEN);
  if (gethostname(lhost, STRLEN - 1) < 0) {
    errmsg("gethostname");
    exit(EX_UNAVAILABLE);
  }
  if (pos = strchr(lhost, '.')) *pos = '\0';

  if (atexit(onexit) < 0) {
    errmsg("atexit");
    return EX_OSERR;
  }
  
  if (cf)
    return udpclt();
  else {
    if (signal(SIGWINCH, SIG_IGN) == SIG_ERR) {
      errmsg("signal");
      return EX_OSERR;
    }
    if (!quiet) printf("%s on %s: listening on port %d\n", pgm, lhost, port);
    return udpsrv();
  }
}
